#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.workers.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     11 Aug 2021, (12:06 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import os
import queue
import shlex
import shutil
import subprocess
import threading
import time

import psutil

from unmanic.libs import common
from unmanic.libs.logs import UnmanicLogging
from unmanic.libs.plugins import PluginsHandler


class WorkerCommandError(Exception):
    def __init___(self, command):
        Exception.__init__(self, "Worker command returned non 0 status. Command: {}".format(command))
        self.command = command


class WorkerSubprocessMonitor(threading.Thread):
    def __init__(self, parent_worker):
        super().__init__(daemon=True)
        self.logger = UnmanicLogging.get_logger(name=__class__.__name__)
        self._stop_event = threading.Event()
        self._terminate_lock = threading.Lock()

        self.parent_worker = parent_worker
        self.event = parent_worker.event
        self.redundant_flag = parent_worker.redundant_flag
        self.paused_flag = parent_worker.paused_flag
        self.paused = False

        # Set current subprocess to None
        self.subprocess_pid = None
        self.subprocess = None
        self.subprocess_start_time = 0
        self.subprocess_pause_time = 0

        # Subprocess stats
        self.subprocess_percent = 0
        self.subprocess_elapsed = 0
        self.subprocess_cpu_percent = 0
        self.subprocess_mem_percent = 0
        self.subprocess_rss_bytes = 0
        self.subprocess_vms_bytes = 0

    def set_proc(self, pid):
        try:
            if pid != self.subprocess_pid:
                self.subprocess_pid = pid
                self.subprocess = psutil.Process(pid=pid)
                # Reset pause time
                self.subprocess_start_time = time.time()
                self.subprocess_pause_time = 0
                # Reset subprocess progress
                self.subprocess_percent = 0
                self.subprocess_elapsed = 0
            if self.redundant_flag.is_set():
                # If the redundant flag is set then we should terminate any set procs straight away as the worker needs to stop
                self.logger.debug("A new subprocess was spawned, but the worker is trying to terminate. Subprocess PID %s",
                                  self.subprocess_pid)
                self.terminate_proc()

        except Exception:
            self.logger.exception("Exception in set_proc()")

    def unset_proc(self):
        try:
            self.subprocess_pid = None
            self.subprocess = None
            # Reset subprocess progress
            self.subprocess_percent = 0
            self.subprocess_elapsed = 0
            # Reset resource values
            self.set_proc_resources_in_parent_worker(0, 0, 0, 0)
        except Exception:
            self.logger.exception("Exception in unset_proc()")

    def set_proc_resources_in_parent_worker(self, normalised_cpu_percent, rss_bytes, vms_bytes, mem_percent):
        self.subprocess_cpu_percent = normalised_cpu_percent
        self.subprocess_rss_bytes = rss_bytes
        self.subprocess_vms_bytes = vms_bytes
        self.subprocess_mem_percent = mem_percent

    def suspend_proc(self):
        # Stop the process if the worker is paused
        # Then resume it when the worker is resumed
        try:
            if not self.subprocess or not self.subprocess.is_running():
                return

            # Create list of all subprocesses - parent + all children
            procs = [self.subprocess] + self.subprocess.children(recursive=True)

            # Suspend them all
            for p in procs:
                try:
                    self.logger.debug("Pausing PID %s", p.pid)
                    p.suspend()
                except psutil.NoSuchProcess:
                    continue

            self.paused = True
            self.subprocess_pause_time = int(time.time())
            while not self.redundant_flag.is_set():
                self.event.wait(1)
                if not self.paused_flag.is_set():
                    # Resume in reverse order
                    for p in reversed(procs):
                        try:
                            self.logger.debug("Resuming PID %s", p.pid)
                            p.resume()
                            # Force anything to shut down straight away if we are exiting the thread
                            if self.redundant_flag.is_set() or self._stop_event.is_set():
                                p.terminate()
                        except psutil.NoSuchProcess:
                            continue
                    self.paused = False
                    break

        except Exception:
            self.logger.exception("Exception in suspend_proc()")

    def terminate_proc(self):
        with self._terminate_lock:
            try:
                # If the process is still running, kill it
                if self.subprocess is not None:
                    self.logger.info("Terminating subprocess PID: %s", self.subprocess_pid)
                    self.__terminate_proc_tree(self.subprocess)
                    self.logger.info("Subprocess terminated")
                    self.unset_proc()
            except Exception:
                self.logger.exception("Exception in terminate_proc()")

    def __log_proc_terminated(self, proc: psutil.Process):
        try:
            self.logger.info("Process %s terminated with exit code %s", proc, proc.returncode)
        except Exception:
            self.logger.exception("Exception in __log_proc_terminated()")

    def __terminate_proc_tree(self, proc: psutil.Process):
        """
        Terminate the process tree (including grandchildren).
        Ensures any suspended processes are first resumed so that
        terminate() will actually take effect.  Processes that
        fail to stop with terminate() within 3s will be killed.

        :param proc:
        :return:
        """
        try:
            # Build the full tree
            all_procs = proc.children(recursive=True) + [proc]

            # Resume all suspended processes so they can handle signals
            for p in all_procs:
                try:
                    p.resume()
                except (psutil.NoSuchProcess, NotImplementedError):
                    pass

            # Attempt graceful shutdown
            for p in all_procs:
                try:
                    p.terminate()
                except psutil.NoSuchProcess:
                    pass

            # Wait up to 3s for them to exit
            gone, alive = psutil.wait_procs(all_procs, timeout=3, callback=self.__log_proc_terminated)

            # Force-kill any remaining processes
            for p in alive:
                try:
                    p.kill()
                except psutil.NoSuchProcess:
                    pass

            # Final wait to reap
            psutil.wait_procs(alive, timeout=3, callback=self.__log_proc_terminated)

        except Exception:
            self.logger.exception("Exception in __terminate_proc_tree()")

    def get_subprocess_elapsed(self):
        try:
            subprocess_elapsed = 0
            if self.subprocess is not None:
                # Get the time now
                now = int(time.time())
                # Get the total running time (including time being paused)
                total_run_time = int(now - self.subprocess_start_time)
                # Get the time when we started being paused
                pause_duration = int(now - self.subprocess_pause_time)
                # Calculate elapsed time of the subprocess subtracting the pause duration
                subprocess_elapsed = int(total_run_time - pause_duration)
            return subprocess_elapsed
        except Exception:
            self.logger.exception("Exception in get_subprocess_elapsed()")
            return 0

    def get_subprocess_stats(self):
        try:
            return {
                'pid':         str(self.subprocess_pid),
                'percent':     str(self.subprocess_percent),
                'elapsed':     self.get_subprocess_elapsed(),
                'cpu_percent': str(self.subprocess_cpu_percent),
                'mem_percent': str(self.subprocess_mem_percent),
                'rss_bytes':   str(self.subprocess_rss_bytes),
                'vms_bytes':   str(self.subprocess_vms_bytes),
            }
        except Exception:
            self.logger.exception("Exception in get_subprocess_stats()")
            # Return something minimal so UI won't break
            return {
                'pid':         '0', 'percent': '0', 'elapsed': '0',
                'cpu_percent': '0', 'mem_percent': '0',
                'rss_bytes':   '0', 'vms_bytes': '0',
            }

    def set_subprocess_start_time(self, proc_start_time):
        try:
            self.subprocess_start_time = proc_start_time
        except Exception:
            self.logger.exception("Exception in set_subprocess_start_time()")

    def set_subprocess_percent(self, percent):
        try:
            self.subprocess_percent = percent
        except Exception:
            self.logger.exception("Exception in set_subprocess_percent()")

    def default_progress_parser(self, line_text, pid=None, proc_start_time=None, unset=False):
        if unset:
            # Here we provide a plugin with the ability to unset a subprocess (indicating that it completed)
            self.unset_proc()
            return {
                'killed':  self.redundant_flag.is_set(),
                'paused':  self.paused,
                'percent': str(self.subprocess_percent),
            }
        try:
            if pid is not None:
                self.set_proc(pid)
            if proc_start_time is not None:
                self.set_subprocess_start_time(proc_start_time)
            try:
                stripped_text = str(line_text).strip()
                text_float = float(stripped_text)
                self.subprocess_percent = int(text_float)
            except (TypeError, ValueError):
                pass
            return {
                'killed':  self.redundant_flag.is_set(),
                'paused':  self.paused,
                'percent': str(self.subprocess_percent),
            }
        except Exception:
            self.logger.exception("Exception in default_progress_parser()")
            return {
                'killed':  self.redundant_flag.is_set(),
                'paused':  self.paused,
                'percent': str(self.subprocess_percent),
            }

    def run(self):
        # First fetch the number of CPUs for normalising the CPU percent
        cpu_count = psutil.cpu_count(logical=True)
        # Loop while thread is expected to be running
        self.logger.warning("Starting WorkerMonitor loop")
        while True:
            try:
                if self._stop_event.is_set():
                    self.event.wait(1)
                    break

                if self.redundant_flag.is_set():
                    # If the worker needs to exit, then terminate the subprocess
                    self.terminate_proc()
                    self.event.wait(1)
                    continue

                if self.subprocess is None:
                    self.event.wait(1)
                    continue

                if not self.subprocess.is_running():
                    self.event.wait(1)
                    continue

                # Fetch CPU info
                cpu_percent = self.subprocess.cpu_percent(interval=None)
                normalised_cpu_percent = cpu_percent / cpu_count

                # Fetch Memory info
                mem_info = self.subprocess.memory_info()
                total_rss = mem_info.rss
                total_vms = mem_info.vms
                for child in self.subprocess.children(recursive=True):
                    try:
                        mem = child.memory_info()
                        total_rss += mem.rss
                        total_vms += mem.vms
                    except psutil.NoSuchProcess:
                        continue

                # Calculate percentage of memory used relative to total system RAM
                total_system_ram = psutil.virtual_memory().total
                mem_percent = (total_rss / total_system_ram) * 100

                # Set values in parent worker thread
                self.set_proc_resources_in_parent_worker(normalised_cpu_percent, total_rss, total_vms, mem_percent)

                # Pause subprocesses if the worker is paused
                if self.paused_flag.is_set():
                    self.suspend_proc()

            except psutil.NoSuchProcess:
                self.logger.debug("No such process: %s", self.subprocess_pid)
            except Exception:
                self.logger.exception("Unhandled exception in WorkerMonitor.run()")

            # Poll interval
            try:
                self.event.wait(1)
            except Exception:
                # In case event.wait itself misbehaves
                self.logger.exception("Exception while waiting in WorkerMonitor.run()")
                time.sleep(1)

        self.logger.info("Exiting WorkerMonitor loop")

    def stop(self):
        self.terminate_proc()
        self._stop_event.set()


class Worker(threading.Thread):
    idle = True
    paused = False

    current_task = None
    worker_log = None
    start_time = None
    finish_time = None

    worker_runners_info = {}

    def __init__(self, thread_id, name, worker_group_id, pending_queue, complete_queue, event):
        super(Worker, self).__init__(name=name)
        self.thread_id = thread_id
        self.name = name
        self.worker_group_id = worker_group_id
        self.event = event

        self.current_task = None
        self.pending_queue = pending_queue
        self.complete_queue = complete_queue
        self.worker_subprocess_monitor = None

        # Create 'redundancy' flag. When this is set, the worker should die
        self.redundant_flag = threading.Event()
        self.redundant_flag.clear()

        # Create 'paused' flag. When this is set, the worker should be paused
        self.paused_flag = threading.Event()
        self.paused_flag.clear()

        # Create logger for this worker
        self.logger = UnmanicLogging.get_logger(name=__class__.__name__)

    def _log(self, message, message2='', level="info"):
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def run(self):
        self.logger.info("Starting worker")

        # Create proc monitor
        self.worker_subprocess_monitor = WorkerSubprocessMonitor(self)
        self.worker_subprocess_monitor.start()

        while not self.redundant_flag.is_set():
            self.event.wait(1)  # Add delay for preventing loop maxing compute resources

            # If the Foreman has paused this worker, then don't do anything
            if self.paused_flag.is_set():
                self.paused = True
                # If the worker is paused, wait for 5 seconds before continuing the loop
                self.event.wait(5)
                continue
            self.paused = False

            # Set the worker as Idle - This will announce to the Foreman that it's ready for a task
            self.idle = True

            # Wait for task
            while not self.redundant_flag.is_set() and self.current_task:
                self.event.wait(.5)  # Add delay for preventing loop maxing compute resources

                try:
                    # Process the set task
                    self.__process_task_queue_item()
                except queue.Empty:
                    continue
                except Exception as e:
                    self._log("Exception in processing job with {}:".format(self.name), message2=str(e),
                              level="exception")

        self.logger.info("Stopping worker")
        self.worker_subprocess_monitor.stop()
        self.worker_subprocess_monitor.join()
        self.worker_subprocess_monitor = None

    def set_task(self, new_task):
        """Sets the given task to the worker class"""
        # Ensure only one task can be set for a worker
        if self.current_task:
            return
        # Set the task
        self.current_task = new_task
        self.worker_log = []
        self.idle = False

    def get_status(self):
        """
        Fetch the status of this worker.

        :return:
        """
        subprocess_stats = None
        if self.worker_subprocess_monitor:
            subprocess_stats = self.worker_subprocess_monitor.get_subprocess_stats()
        status = {
            'id':              str(self.thread_id),
            'name':            self.name,
            'idle':            self.idle,
            'paused':          self.paused_flag.is_set(),
            'start_time':      None if not self.start_time else str(self.start_time),
            'current_task':    None,
            'current_file':    "",
            'worker_log_tail': [],
            'runners_info':    {},
            'subprocess':      subprocess_stats,
        }
        if self.current_task:
            # Fetch the current file
            try:
                status['current_task'] = self.current_task.get_task_id()
            except Exception as e:
                self._log("Exception in fetching the current task ID for worker {}:".format(self.name), message2=str(e),
                          level="exception")

            # Fetch the current file
            try:
                status['current_file'] = self.current_task.get_source_basename()
            except Exception as e:
                self._log("Exception in fetching the current file of worker {}:".format(self.name), message2=str(e),
                          level="exception")

            # Append the worker log tail
            try:
                if self.worker_log and len(self.worker_log) > 20:
                    status['worker_log_tail'] = self.worker_log[-19:]
                else:
                    status['worker_log_tail'] = self.worker_log
            except Exception as e:
                self._log("Exception in fetching log tail of worker: ", message2=str(e),
                          level="exception")

            # Append the runners info
            try:
                status['runners_info'] = self.worker_runners_info
            except Exception as e:
                self._log("Exception in runners info of worker {}:".format(self.name), message2=str(e),
                          level="exception")
        return status

    def __unset_current_task(self):
        self.current_task = None
        self.worker_runners_info = {}
        self.worker_log = []

    def __process_task_queue_item(self):
        """
        Processes the set task.

        :return:
        """
        # Mark worker as not idle now that it is processing a task
        self.idle = False

        # Log the start of the job
        self.logger.info("Picked up job - %s", self.current_task.get_source_abspath())

        # Mark as being "in progress"
        self.current_task.set_status('in_progress')

        # Start current task stats
        self.__set_start_task_stats()

        # Process the file. Will return true if success, otherwise false
        success = self.__exec_worker_runners_on_set_task()
        # Mark the task as either success or not
        self.current_task.set_success(success)

        # Mark task completion statistics
        self.__set_finish_task_stats()

        # Log completion of job
        self.logger.info("Finished job - %s", self.current_task.get_source_abspath())

        # Place the task into the completed queue
        self.complete_queue.put(self.current_task)

        # Reset the current file info for the next task
        self.__unset_current_task()

    def __set_start_task_stats(self):
        """Sets the initial stats for the start of a task"""
        # Set the start time to now
        self.start_time = time.time()

        # Clear the finish time
        self.finish_time = None

        # Format our starting statistics data
        self.current_task.task.processed_by_worker = self.name
        self.current_task.task.start_time = self.start_time
        self.current_task.task.finish_time = self.finish_time

    def __set_finish_task_stats(self):
        """Sets the final stats for the end of a task"""
        # Set the finish time to now
        self.finish_time = time.time()

        # Set the finish time in the statistics data
        self.current_task.task.finish_time = self.finish_time

    def __exec_worker_runners_on_set_task(self):
        """
        Executes the configured plugin runners against the set task.

        :return:
        """
        # Init plugins
        library_id = self.current_task.get_task_library_id()
        plugin_handler = PluginsHandler()
        plugin_modules = plugin_handler.get_enabled_plugin_modules_by_type('worker.process', library_id=library_id)

        # Create dictionary of runners info for the frontend
        self.worker_runners_info = {}
        for plugin_module in plugin_modules:
            self.worker_runners_info[plugin_module.get('plugin_id')] = {
                'plugin_id':   plugin_module.get('plugin_id'),
                'status':      'pending',
                "name":        plugin_module.get('name'),
                "author":      plugin_module.get('author'),
                "version":     plugin_module.get('version'),
                "icon":        plugin_module.get('icon'),
                "description": plugin_module.get('description'),
            }

        # Set the absolute path to the original file
        original_abspath = self.current_task.get_source_abspath()

        # Process item in loop.
        # First process the item for each plugin that configures it, then run the default Unmanic configuration
        task_cache_path = self.current_task.get_cache_path()
        cache_directory = os.path.dirname(os.path.abspath(task_cache_path))
        # Set the current input file to the original file path
        file_in = original_abspath
        # Mark the overall success of all runners. This will be set to False if any of the runners fails.
        overall_success = True
        # Set the current file out to nothing.
        # This will be configured by each runner.
        # If no runners are configured, then nothing needs to be done.
        current_file_out = original_abspath
        # The number of runners that have been run
        runner_count = 0
        # Flag if a task has run a command
        no_exec_command_run = True

        # Execute event plugin runners
        plugin_handler.run_event_plugins_for_plugin_type('events.worker_process_started', {
            "library_id":          library_id,
            "task_type":           self.current_task.get_task_type(),
            "original_file_path":  original_abspath,
            "cache_directory":     cache_directory,
            "worker_runners_info": self.worker_runners_info,
        })

        # Generate default data object for the runner functions
        task_id = self.current_task.get_task_id()
        data = {
            "worker_log":              self.worker_log,
            "library_id":              library_id,
            "exec_command":            [],
            "command_progress_parser": None,
            "file_in":                 file_in,
            "file_out":                None,
            "original_file_path":      original_abspath,
            "repeat":                  False,
        }

        for plugin_module in plugin_modules:
            # Increment the runners count (first runner will be set as #1)
            runner_count += 1
            runner_id = plugin_module.get('plugin_id')

            if not overall_success:
                # If one of the Plugins fails, don't continue.
                # The Plugins could be co-dependant and the final file will not go anywhere if 'overall_success' is False
                break

            # Mark the status of the worker for the frontend
            self.worker_runners_info[runner_id]['status'] = 'in_progress'
            self.worker_runners_info[runner_id]['success'] = False

            # Loop over runner. This way we can repeat the function with the same data if requested by the repeat flag
            runner_pass_count = 0
            while not self.redundant_flag.is_set():
                runner_pass_count += 1

                # Fetch file out details
                # This creates a temp file labeled "WORKING" that will be moved to the cache_path on completion
                split_file_out = os.path.splitext(task_cache_path)
                split_file_in = os.path.splitext(file_in)
                file_out = "{}-{}-{}-{}{}".format(split_file_out[0], "WORKING", runner_count, runner_pass_count,
                                                  split_file_in[1])

                # Reset data object for this runner functions
                data['library_id'] = library_id
                data['exec_command'] = []
                data['command_progress_parser'] = self.worker_subprocess_monitor.default_progress_parser
                data['file_in'] = file_in
                data['file_out'] = file_out
                data['original_file_path'] = original_abspath
                data['repeat'] = False
                data['task_id'] = task_id

                self.event.wait(.2)  # Add delay for preventing loop maxing compute resources
                self.worker_log.append("\n\nRUNNER: \n{} [Pass #{}]\n\n".format(plugin_module.get('name'), runner_pass_count))
                self.worker_log.append("\nExecuting plugin runner... Please wait\n")

                # Run plugin (in its own thread) to update data
                result = {"success": None}

                def _run_plugin():
                    result["success"] = plugin_handler.exec_plugin_runner(
                        data, runner_id, 'worker.process'
                    )

                runner_thread = threading.Thread(target=_run_plugin, daemon=True)
                runner_thread.start()

                # monitor the thread, bail if redundancy requested
                while runner_thread.is_alive():
                    if self.redundant_flag.is_set():
                        self.logger.warning("Worker stop flag set, aborting plugin runner '%s'", runner_id)
                        break
                    self.event.wait(0.2)

                # if we were told to shut down, mark failure and exit loop
                if self.redundant_flag.is_set():
                    self.worker_runners_info[runner_id]['status'] = 'complete'
                    self.worker_runners_info[runner_id]['success'] = False
                    overall_success = False
                    self.worker_log.append("\n\nWORKER TERMINATED!")
                    break

                # now check the plugin result
                if not result["success"]:
                    # Skip this plugin module's loop
                    self.worker_runners_info[runner_id]['status'] = 'complete'
                    self.worker_runners_info[runner_id]['success'] = False
                    # Set overall success status to failed
                    overall_success = False
                    # Append long entry to say the worker was terminated
                    self.worker_log.append("\n\nPLUGIN FAILED!")
                    self.worker_log.append("\nFailed to execute Plugin '{}'".format(plugin_module.get('name')))
                    self.worker_log.append("\nCheck Unmanic logs for more information")
                    break

                # Log the in and out files returned by the plugin runner for debugging
                self._log("Worker process '{}' (in)".format(runner_id), data.get("file_in"),
                          level='debug')
                self._log("Worker process '{}' (out)".format(runner_id), data.get("file_out"),
                          level='debug')

                # Only run the conversion process if "exec_command" is not empty
                if data.get("exec_command"):
                    self.worker_log.append("\nPlugin runner requested for a command to be executed by Unmanic")

                    # Exec command as subprocess
                    success = self.__exec_command_subprocess(data)
                    no_exec_command_run = False

                    if self.redundant_flag.is_set():
                        # This worker has been marked as redundant. It is being terminated.
                        self._log("Worker has been terminated before a command was completed", level="warning")
                        # Mark runner as failed
                        self.worker_runners_info[runner_id]['success'] = False
                        # Set overall success status to failed
                        overall_success = False
                        # Append long entry to say the worker was terminated
                        self.worker_log.append("\n\nWORKER TERMINATED!")
                        # Don't continue
                        break

                    # Check if command exited successfully.
                    if success:
                        # If file conversion was successful
                        self.logger.info("Successfully ran worker process '%s' on file '%s'",
                                         runner_id,
                                         data.get("file_in"))
                        # Check if 'file_out' was nulled by the plugin. If it is, then we will assume that the plugin modified the file_in in-place
                        if not data.get('file_out'):
                            # The 'file_out' is None. Ensure the new 'file_in' is set to whatever the plugin returned for 'file_in' for the next loop
                            file_in = data.get("file_in")
                        # Ensure the 'file_out' that was specified by the plugin to be created was actually created.
                        elif os.path.exists(data.get('file_out')):
                            # The outfile exists...
                            # In order to clean up as we go and avoid unnecessary RAM/disk use in the cache directory,
                            #   we want to remove the 'file_in' file.
                            # We want to ensure that we do not accidentally remove any original files here.
                            # We also want to ensure that the 'file_out' is not removed if the plugin set it to the same path as the 'file_in'.
                            # To avoid this, run x3 tests.
                            # First, check current 'file_in' is not the original file.
                            if os.path.abspath(data.get("file_in")) != os.path.abspath(original_abspath):
                                # Second, check that the 'file_in' is actually in cache directory. If it is not, we did not create it.
                                if "unmanic_file_conversion" in os.path.abspath(data.get("file_in")):
                                    # Finally, check that the file_out is not the same file as the file_in
                                    if os.path.abspath(data.get("file_out")) != os.path.abspath(data.get("file_in")):
                                        # Remove the old file_in file
                                        os.remove(os.path.abspath(data.get("file_in")))

                            # Set the new 'file_in' as the previous runner's 'file_out' for the next loop
                            file_in = data.get("file_out")
                    else:
                        # If file conversion was not successful
                        self._log(
                            "Error while running worker process '{}' on file '{}'".format(
                                runner_id,
                                original_abspath
                            ),
                            level="error")
                        self.worker_runners_info[runner_id]['success'] = False
                        overall_success = False
                else:
                    # Ensure the new 'file_in' is set to the previous runner's 'file_in' for the next loop
                    file_in = data.get("file_in")
                    # Log that this plugin did not request to execute anything
                    self.worker_log.append("\nRunner did not request for Unmanic to execute a command")
                    self._log(
                        "Worker process '{}' did not request to execute a command.".format(runner_id),
                        level='debug')

                if data.get('file_out') and os.path.exists(data.get('file_out')):
                    # Set the current file out to the most recently completed cache file
                    # If the file out does not exist, it is likely never used by the plugin.
                    current_file_out = data.get('file_out')
                else:
                    # Ensure the current_file_out is set the currently set 'file_in'
                    current_file_out = data.get('file_in')

                if data.get("repeat"):
                    # The returned data contained the 'repeat'' flag.
                    # Run another pass against this same plugin
                    continue
                break

            self.worker_runners_info[runner_id]['success'] = True
            self.worker_runners_info[runner_id]['status'] = 'complete'

        # Log if no command was run by any Plugins
        if no_exec_command_run:
            # If no jobs were carried out on this task
            self._log("No Plugin requested for Unmanic to run commands for this file '{}'".format(original_abspath),
                      level='warning')
            self.worker_log.append(
                "\n\nNo Plugin requested for Unmanic to run commands for this file '{}'".format(original_abspath))

        # Save the completed command log
        self.current_task.save_command_log(self.worker_log)

        # If all plugins that were executed completed successfully, then this was overall a successful task.
        # At this point we need to move the final out file to the original task cache path so the postprocessor can collect it.
        if overall_success:
            # If jobs carried out on this task were all successful, we will get here
            self.logger.info("Successfully completed Worker processing on file '%s'", original_abspath)

            # Attempt to move the final output file to the final cache file path for the postprocessor
            try:
                # Set the new file out as the extension may have changed
                split_file_name = os.path.splitext(current_file_out)
                file_extension = split_file_name[1].lstrip('.')
                self.current_task.set_cache_path(cache_directory, file_extension)
                # Read the updated cache path
                task_cache_path = self.current_task.get_cache_path()

                # Move file to original cache path
                self._log("Moving final cache file from '{}' to '{}'".format(current_file_out, task_cache_path))
                current_file_out = os.path.abspath(current_file_out)

                # There is a really odd intermittent bug with the shutil module that is causing it to
                #   sometimes report that the file does not exist.
                # This section adds a small pause and logs the error if that is the case.
                # I have not yet figured out a solution as this is difficult to reproduce.
                if not os.path.exists(current_file_out):
                    self._log("Error - current_file_out path does not exist! '{}'".format(file_in), level="error")
                    self.event.wait(1)

                # Ensure the cache directory exists
                if not os.path.exists(cache_directory):
                    os.makedirs(cache_directory)

                # Check that the current file out is not the original source file
                if os.path.abspath(current_file_out) == os.path.abspath(original_abspath):
                    # The current file out is not a cache file, the file must have never been modified.
                    # This can happen if all Plugins failed to run, or a Plugin specifically reset the out
                    #   file to the original source in order to preserve it.
                    # In this circumstance, we want to create a cache copy and let the process continue.
                    self._log("Final cache file is the same path as the original source. Creating cache copy.", level='debug')
                    shutil.copyfile(current_file_out, task_cache_path)
                else:
                    # Use shutil module to move the file to the final task cache location
                    shutil.move(current_file_out, task_cache_path)
            except Exception as e:
                self._log("Exception in final move operation of file {} to {}:".format(current_file_out, task_cache_path),
                          message2=str(e), level="exception")
                overall_success = False

        # Execute event plugin runners (only when added to queue)
        plugin_handler.run_event_plugins_for_plugin_type('events.worker_process_complete', {
            "library_id":          library_id,
            "task_type":           self.current_task.get_task_type(),
            "original_file_path":  original_abspath,
            "final_cache_path":    task_cache_path,
            "overall_success":     overall_success,
            "worker_runners_info": self.worker_runners_info,
            "worker_log":          self.worker_log,
        })

        # If the overall result of the jobs carried out on this task were not successful, log the failure and return False
        if not overall_success:
            self._log("Failed to process task for file '{}'".format(original_abspath), level='warning')
        return overall_success

    def __exec_command_subprocess(self, data):
        """
        Executes a command as a shell subprocess.
        Uses the given parser to record progress data from the shell STDOUT.

        :param data:
        :return:
        """
        # Fetch command to execute.
        exec_command = data.get("exec_command", [])

        # Fetch the command progress parser function
        command_progress_parser = data.get("command_progress_parser", self.worker_subprocess_monitor.default_progress_parser)

        # Log the command for debugging
        command_string = exec_command
        if isinstance(exec_command, list):
            command_string = shlex.join(exec_command)
        self._log("Executing: {}".format(command_string), level='debug')

        # Append start of command to worker subprocess stdout
        self.worker_log += [
            '\n\n',
            'COMMAND:\n',
            command_string,
            '\n\n',
            'LOG:\n',
        ]

        # Create output path if file_out is present and the path does not exists
        if data.get("file_out"):
            common.ensure_dir(data.get("file_out"))

        # Convert file
        try:
            # Execute command
            if isinstance(exec_command, list):
                sub_proc = subprocess.Popen(exec_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                            universal_newlines=True, errors='replace')
            elif isinstance(exec_command, str):
                sub_proc = subprocess.Popen(exec_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                            universal_newlines=True, errors='replace', shell=True)
            else:
                raise Exception(
                    "Plugin's returned 'exec_command' object must be either a list or a string. Received type {}.".format(
                        type(exec_command)))

            # Fetch process using psutil for control (sending SIGSTOP on windows will not work)
            proc = psutil.Process(pid=sub_proc.pid)

            # Create proc monitor
            self.worker_subprocess_monitor.set_proc(sub_proc.pid)

            # Set process priority on posix systems
            # TODO: Test how this will work on Windows
            if os.name == "posix":
                try:
                    parent_proc = psutil.Process(os.getpid())
                    parent_proc_nice = parent_proc.nice()
                    proc.nice(parent_proc_nice + 1)
                except Exception as e:
                    self._log("Unable to lower priority of subprocess. Subprocess should continue to run at normal priority",
                              str(e), level='warning')

            # Poll process for new output until finished
            while not self.redundant_flag.is_set():

                # Stop parsing the sub process if the worker is paused
                # Then resume it when the worker is resumed
                if self.paused_flag.is_set():
                    self.logger.debug("Pausing worker exec command subprocess loop")
                    while not self.redundant_flag.is_set():
                        self.event.wait(1)
                        if not self.paused_flag.is_set():
                            self.logger.debug("Resuming worker exec command subprocess loop")
                            break
                        continue

                # Fetch command stdout and append it to the current task object (to be saved during post process)
                line_text = sub_proc.stdout.readline()
                self.worker_log.append(line_text)

                # Check if the command has completed. If it has, exit the loop
                if line_text == '' and sub_proc.poll() is not None:
                    self._log("Subprocess task completed!", level='debug')
                    break

                # Parse the progress
                try:
                    progress_dict = command_progress_parser(line_text)
                    progress_percent = progress_dict.get('percent', 0)
                    self.worker_subprocess_monitor.set_subprocess_percent(progress_percent)
                except Exception as e:
                    # Only need to show any sort of exception if we have debugging enabled.
                    # So we should log it as a debug rather than an exception.
                    self._log("Exception while parsing command progress", str(e), level='debug')

            # Get the final output and the exit status
            if not self.redundant_flag.is_set():
                communicate = sub_proc.communicate()[0]

            # If the process is still running, kill it
            self.worker_subprocess_monitor.terminate_proc()

            # Stop proc monitor
            self.worker_subprocess_monitor.unset_proc()

            if sub_proc.returncode == 0:
                return True
            else:
                self._log("Command run against '{}' exited with non-zero status. "
                          "Download command dump from history for more information.".format(data.get("file_in")),
                          message2=str(exec_command), level="error")
                return False

        except Exception as e:
            self._log("Error while executing the command against file{}.".format(data.get("file_in")), message2=str(e),
                      level="error")

        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.__init__.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     05 Mar 2021, (8:12 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

from __future__ import absolute_import

import os
from importlib import import_module
from pathlib import Path
import sys
import inspect
import pkgutil

from .plugin_type_base import PluginType

d = os.path.join(Path(__file__).parent)
type_modules_paths = [os.path.join(d, o) for o in os.listdir(d) if os.path.isdir(os.path.join(d, o)) and not o.startswith('_')]


def grab_module(module_name, *args, **kwargs):
    """
    Fetch a module by name and return the instance of that class

    :param module_name:
    :param args:
    :param kwargs:
    :return:
    """
    try:
        if '.' in module_name:
            module_name, class_name = module_name.rsplit('.', 1)
        else:
            class_name = module_name.capitalize()

        module = import_module('.' + module_name, package=__name__)
        module_class = getattr(module, class_name)
        instance = module_class(*args, **kwargs)

        return instance

    except (AttributeError, AssertionError, ModuleNotFoundError):
        raise ImportError('{} is not part of our supported plugin types!'.format(module_name))


def get_all_plugin_types():
    """
    Fetch a list of supported plugin types and
    return a dictionary of their data

    :return:
    """
    return_dic = {}

    for type_modules_path in type_modules_paths:
        plugin_type = os.path.basename(Path(type_modules_path))
        for (_, module_name, _) in pkgutil.iter_modules([type_modules_path]):
            instance = grab_module(plugin_type + '.' + module_name)
            return_data = {
                "name": instance.plugin_type_name(),
                "runner": instance.plugin_runner(),
                "runner_docstring": instance.plugin_runner_docstring(),
            }
            return_dic[plugin_type + '.' + module_name] = return_data

    return return_dic


"""
Import all submodules for this package

"""
for type_modules_path in type_modules_paths:
    plugin_type = os.path.basename(Path(type_modules_path))
    for (_, name, _) in pkgutil.iter_modules([type_modules_path]):

        imported_module = import_module('.' + plugin_type + '.' + name, package=__name__)

        for i in dir(imported_module):
            attribute = getattr(imported_module, i)

            if inspect.isclass(attribute) and issubclass(attribute, PluginType):
                setattr(sys.modules[__name__], name, attribute)

__author__ = 'Josh.5 (jsunnex@gmail.com)'

__all__ = (
    'PluginType',
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.unlogger.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     06 Jan 2019, (8:41 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

import os
import logging
import threading
import json
import time
from logging.handlers import RotatingFileHandler
from queue import Queue, Empty
import requests
from datetime import datetime, timedelta
from json_log_formatter import JSONFormatter


class ForwardJSONFormatter(JSONFormatter):
    """
    JSON log formatter which adds log record attributes if debugging is enabled.
    """

    def json_record(self, message, extra, record):
        # Check the logger's effective level
        logger = logging.getLogger(record.name)
        # Always include levelname used for labels
        extra['levelname'] = record.levelname
        # If the logger's effective level is DEBUG, add more context
        if logger.getEffectiveLevel() == logging.DEBUG:
            extra['filename'] = record.filename
            extra['funcName'] = record.funcName
            extra['lineno'] = record.lineno
            extra['module'] = record.module
            extra['name'] = record.name
            extra['pathname'] = record.pathname
            extra['process'] = record.process
            extra['processName'] = record.processName
            if hasattr(record, 'stack_info'):
                extra['stack_info'] = record.stack_info
            else:
                extra['stack_info'] = None
            extra['thread'] = record.thread
            extra['threadName'] = record.threadName
        # Choose time from metric_timestamp or data_timestamp
        ts_str = extra.get('metric_timestamp') or extra.get('data_timestamp')
        if ts_str:
            try:
                ts_float = float(ts_str)
                extra['time'] = datetime.utcfromtimestamp(ts_float).isoformat()
            except Exception:
                pass  # Ignore this. The default formatter will add a "time" record
        return super(ForwardJSONFormatter, self).json_record(message, extra, record)


class ForwardLogHandler(logging.Handler):
    """
    A custom log handler that:
    - Writes incoming logs to disk in chunk files at intervals or size thresholds.
    - A separate thread reads these chunk files from disk and forwards them.
    - On success, chunk files are removed.
    - On failure, chunk files remain for later retry.
    """

    def __init__(self, buffer_path, installation_name, labels=None, flush_interval=5, max_chunk_size=5 * 1024 * 1024):
        # TODO: Set default flush interval to 10 seconds
        super().__init__()
        self.buffer_path = buffer_path
        self.endpoint = None
        self.app_id = None
        self.installation_name = installation_name
        self.labels = labels if labels is not None else {"job": "unmanic"}
        self.flush_interval = flush_interval
        self.max_chunk_size = max_chunk_size

        self.log_queue = Queue()
        self.stop_event = threading.Event()

        # Thread that processes logs from the queue and writes them to disk
        self.writer_thread = threading.Thread(target=self._process_logs, daemon=True)
        self.writer_thread.start()

        # Thread that periodically attempts to send logs from disk
        self.sender_thread = threading.Thread(target=self._send_from_disk, daemon=True)
        self.sender_thread.start()

    def configure_endpoint(self, endpoint, app_id):
        self.endpoint = endpoint
        self.app_id = app_id

    def emit(self, record):
        try:
            log_entry = self.format(record)

            # Set log timestamp in nanoseconds
            ts = str(int(time.time() * 1e9))
            if hasattr(record, 'created'):
                ts = str(int(record.created * 1e9))

            # Set default labels
            labels = {
                "service_name":      "unmanic",  # This is a required label
                "logger":            record.name,
                "level":             record.levelname,
                "installation_name": self.installation_name,
                "log_type":          "APPLICATION_LOG",
            }

            # If the record has a log_type attribute, override
            if hasattr(record, 'log_type') and record.log_type:
                labels['log_type'] = record.log_type

            # If the record has a metric_name attribute, add it as a label
            if hasattr(record, 'metric_name') and record.metric_name:
                labels['metric_name'] = record.metric_name

            # If the record has a data_primary_key attribute, add it as a label
            if hasattr(record, 'data_primary_key') and record.data_primary_key:
                labels['data_primary_key'] = record.data_primary_key

            self.log_queue.put({
                "labels": labels,
                "entry":  [ts, log_entry]
            })
        except Exception as e:
            logging.getLogger("Unmanic.ForwardLogHandler").error("Failed to enqueue log: %s", e)

    def _process_logs(self):
        """
        Reads logs from the queue and writes them to disk in chunks.
        Chunks are created either at flush_interval timeout or if max_chunk_size is reached.
        """
        buffer = []
        buffer_size = 0
        last_flush_time = int(time.time())

        while not self.stop_event.is_set():

            try:
                log_entry = self.log_queue.get(timeout=0.2)
                if log_entry is None:
                    # Sentinel received, means shutdown is requested
                    break
                # Process the log entry
                buffer.append(log_entry)
                buffer_size += len(log_entry)

                # If size exceeds max_chunk_size, flush now
                if buffer_size >= self.max_chunk_size:
                    self._flush_to_disk(buffer)
                    buffer = []
                    buffer_size = 0
                    last_flush_time = int(time.time())

            except Empty:
                # No logs available right now
                pass

            current_time = int(time.time())
            if current_time - last_flush_time >= 5:
                # Flush to disk if more than 5 seconds have passed
                if buffer:
                    self._flush_to_disk(buffer)
                    buffer = []
                    buffer_size = 0
                    last_flush_time = current_time

        # Outside the loop, if we are shutting down and have pending logs, flush them
        if buffer:
            self._flush_to_disk(buffer)

    def _flush_to_disk(self, buffer):
        """
        Write the given buffer of logs to a new chunk file on disk.
        The file name includes a timestamp to ensure ordering by time.
        """
        try:
            if not os.path.exists(self.buffer_path):
                os.makedirs(self.buffer_path)

            if not buffer:
                # Nothing in buffer
                return

            timestamp_str = datetime.utcnow().isoformat()
            buffer_file = os.path.join(self.buffer_path, f"log_buffer_{timestamp_str}.json.lock")

            with open(buffer_file, "w") as f:
                for log_entry in buffer:
                    f.write(json.dumps(log_entry) + "\n")

            # Rename the file to indicate it is ready for processing
            os.rename(buffer_file, buffer_file.rstrip(".lock"))
        except Exception as e:
            logging.getLogger("Unmanic.ForwardLogHandler").error("Failed to save logs to disk: %s", e)

    def _send_from_disk(self):
        """
        Periodically attempts to send logs from disk.
        Processes one file at a time, oldest first (important).
        If successful (204), deletes the file.
        If failed, leaves the file for later retry.
        """
        while not self.stop_event.is_set():
            # Just loop if no endpoint is set
            if not self.endpoint or not self.app_id:
                self.stop_event.wait(timeout=10)
                continue
            # Attempt to send the oldest file first
            buffer_files = self._get_buffer_files()
            if not buffer_files:
                # No files to send, sleep
                self.stop_event.wait(timeout=10)
                continue

            for buffer_file in buffer_files:
                # Ignore files that are too old
                if self._buffer_file_too_old(buffer_file):
                    os.remove(buffer_file)
                    self.stop_event.wait(timeout=0.2)
                    continue
                # Ignore if no endpoint is set
                if not self.endpoint or not self.app_id:
                    self.stop_event.wait(timeout=0.2)
                    continue
                if not self._attempt_send_file(buffer_file):
                    # Add a longer pause here and then retry after the delay
                    self.stop_event.wait(timeout=60)
                self.stop_event.wait(timeout=0.2)
                if self.stop_event.is_set():
                    break
            self.stop_event.wait(timeout=0.2)

    def _get_buffer_files(self):
        """
        Returns a sorted list of buffer files based on their timestamp in the filename.
        """
        if not os.path.exists(self.buffer_path):
            return []
        files = [f for f in os.listdir(self.buffer_path) if f.startswith("log_buffer_") and f.endswith(".json")]
        files.sort()
        return [os.path.join(self.buffer_path, f) for f in files]

    def _attempt_send_file(self, buffer_file):
        """
        Attempt to send the logs from buffer_file.
        On success, delete the file.
        On failure, leave it for next retry.
        """
        try:
            buffer = []
            with open(buffer_file, "r") as f:
                for line in f:
                    line = line.strip()
                    if line:
                        buffer.append(json.loads(line))

            if not buffer:
                # Empty file, just remove it
                os.remove(buffer_file)
                return True

            payload = self._create_payload(buffer)
            response = requests.post(f"{self.endpoint}/api/v1/push", json=payload,
                                     headers={"Content-Type": "application/json"})

            if response.status_code == 204:
                # Success, remove the file
                os.remove(buffer_file)
                return True
            else:
                logging.getLogger("Unmanic.ForwardLogHandler").error("Failed to forward logs from %s to remote host %s: %s %s",
                                                                     os.path.basename(buffer_file),
                                                                     self.endpoint,
                                                                     response.status_code,
                                                                     response.text)
                # Leave the file for later retry
        except requests.exceptions.ConnectionError as e:
            # Ignore this. We will try again later
            logging.getLogger("Unmanic.ForwardLogHandler").warning(
                "ConnectionError on remote endpoint %s. Ensure this URL is reachable by Unmanic.",
                self.endpoint)
        except Exception as e:
            logging.getLogger("Unmanic.ForwardLogHandler").exception("Exception while trying to forward logs from %s: %s",
                                                                     buffer_file, e)
            # Leave the file for later retry
        return False

    @staticmethod
    def _buffer_file_too_old(buffer_file, max_days=14):
        """
        Check if the log buffer file is older than the specified number of days based on its timestamp in the filename.

        The expected filename format is:
          log_buffer_YYYY-MM-DDTHH:MM:SS.ffffff.json

        Returns True if the timestamp is older than max_days, otherwise False.
        """
        basename = os.path.basename(buffer_file)
        prefix = "log_buffer_"
        suffix = ".json"

        if not (basename.startswith(prefix) and basename.endswith(suffix)):
            return False  # Filename doesn't match expected pattern.

        # Extract the timestamp part from the filename.
        timestamp_str = basename[len(prefix):-len(suffix)]

        try:
            file_timestamp = datetime.fromisoformat(timestamp_str)
        except ValueError:
            # Unable to parse the timestamp.
            return False

        max_age_threshold = datetime.now() - timedelta(days=max_days)
        return file_timestamp < max_age_threshold

    def _create_payload(self, buffer):
        """
        Create the payload from the given buffer of logs.
        {
          "streams": [
            {
              "stream": { "label": "value", ... },
              "values": [
                [ "<ts in ns>", "<log line>" ],
                ...
              ]
            }
          ]
        }
        """
        combined_streams = {}
        for log_item in buffer:
            stream_key = frozenset(log_item["labels"].items())
            if stream_key not in combined_streams:
                combined_streams[stream_key] = {
                    "stream": dict(log_item["labels"]),
                    "values": []
                }
            combined_streams[stream_key]["values"].append(log_item["entry"])

        return {
            "app_id": self.app_id,
            "data":   {"streams": list(combined_streams.values())}
        }

    def close(self):
        """
        Stop the log handler gracefully.
        Ensures all logs in the queue are flushed to disk.
        """
        # Place Sentinel entry into queue to indicate that no more logs should be processed
        self.log_queue.put(None)

        # Signal the thread to stop
        self.stop_event.set()

        # Wait for the threads to finish processing
        self.writer_thread.join()
        self.sender_thread.join()

        # Explicitly flush any remaining logs in the queue
        remaining_logs = []
        while True:
            try:
                log_entry = self.log_queue.get(timeout=1)
                if log_entry is not None:
                    remaining_logs.append(log_entry)
            except Empty:
                # No logs available right now, break loop
                break

        if remaining_logs:
            self._flush_to_disk(remaining_logs)

        super().close()


class UnmanicLogging:
    METRIC = 9
    DATA = 8
    _instance = None
    _lock = threading.Lock()
    _configured = False
    _log_path = None
    stream_handler = None  # Stream handler
    file_handler = None  # File handler
    remote_handler = None  # Remote log handler

    def __new__(cls):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super(UnmanicLogging, cls).__new__(cls)
                cls._instance._logger = logging.getLogger("Unmanic")
                logging.addLevelName(cls._instance.METRIC, "METRIC")
                cls._instance._logger.setLevel(logging.INFO)
                cls._instance._logger.propagate = False
            return cls._instance

    @staticmethod
    def get_logger(name=None, settings=None):
        """
        Get a child logger. Configure the root logger if 'settings' are provided.
        """
        logger_instance = UnmanicLogging()
        if settings and not logger_instance._configured:
            logger_instance.configure(settings)

        if name:
            return logging.getLogger(f"Unmanic.{name}")
        return logger_instance._logger

    def configure(self, settings):
        """
        Configure the logger using the provided Config settings instance.

        :param settings: Instance of Config class with application settings.
        """
        with self._lock:
            if self._configured:
                return
            # Get logger for this class
            init_logger = logging.getLogger(f"Unmanic.UnmanicLogging")

            # Default formatter
            formatter = logging.Formatter(
                '%(asctime)s:%(levelname)s:%(name)s - %(message)s',
                datefmt='%Y-%m-%dT%H:%M:%S'
            )

            # Set up stream handler
            if self.stream_handler is None:
                self.stream_handler = logging.StreamHandler()
                self.stream_handler.setFormatter(formatter)
                # Set the log level of the stream handle for this log line only
                self.stream_handler.setLevel(logging.INFO)
                self._logger.addHandler(self.stream_handler)
                # Add an info log to let users know where to look for their logs
                init_logger.info("Initialising file logger. All further logs should output to the 'unmanic.log' file")
                # Set the log level of the stream handle always to error
                self.stream_handler.setLevel(logging.CRITICAL)

            # Set up file handler if log path exists
            self._log_path = settings.get_log_path()
            if self._log_path:
                if not os.path.exists(self._log_path):
                    os.makedirs(self._log_path)

                self.file_handler = RotatingFileHandler(
                    os.path.join(self._log_path, "unmanic.log"), maxBytes=10 * 1024 * 1024, backupCount=5
                )
                self.file_handler.setFormatter(formatter)
                # Set file handler log level based on debugging setting
                self.file_handler.setLevel(logging.DEBUG if settings.get_debugging() else logging.INFO)
                self._logger.addHandler(self.file_handler)

            # Setup ForwardLogHandler
            json_formatter = ForwardJSONFormatter()
            buffer_path = os.path.join(self._log_path, "buffer")
            installation_name = settings.get_installation_name()
            self.remote_handler = ForwardLogHandler(buffer_path, installation_name)
            self.remote_handler.setFormatter(json_formatter)
            self.remote_handler.setLevel(self.DATA)
            self._logger.addHandler(self.remote_handler)

            # Set root logger level
            self._logger.setLevel(self.DATA)
            self._configured = True

    @staticmethod
    def metric(name: str, timestamp: datetime = None, **kwargs):
        """
        Custom log method for the METRIC level.
        Logs directly to the remote_handler, if enabled.
        """
        instance = UnmanicLogging()
        if not timestamp:
            timestamp = datetime.now()
        log_record = {
            'log_type':         'METRIC',
            'metric_name':      name,
            'metric_timestamp': f"{int(timestamp.timestamp())}.{timestamp.microsecond}",
            **kwargs
        }
        log_message = " ".join(
            f'{key}="{value}"' if " " in str(value) else f"{key}={value}"
            for key, value in log_record.items() if value
        )
        instance._logger.log(instance.METRIC, log_message, extra=log_record)

    @staticmethod
    def data(data_primary_key: str, data_search_key: str = None, timestamp: datetime = None, **kwargs):
        """
        Custom log method for the DATA level.
        Logs directly to the remote_handler, if enabled.
        """
        instance = UnmanicLogging()
        if not timestamp:
            timestamp = datetime.now()
        log_record = {
            'log_type':         'DATA',
            'data_primary_key': data_primary_key,
            'data_search_key':  data_search_key,
            'data_timestamp':   f"{int(timestamp.timestamp())}.{timestamp.microsecond}",
            **kwargs
        }
        log_message = "DATA STREAM"
        instance._logger.log(instance.DATA, log_message, extra=log_record)

    @staticmethod
    def enable_debugging():
        """
        Enable debugging globally across all threads.
        """
        instance = UnmanicLogging()
        instance._logger.setLevel(logging.DEBUG)
        instance._logger.info("Log level set to DEBUG")

    @staticmethod
    def disable_debugging():
        """
        Disable debugging globally across all threads.
        """
        instance = UnmanicLogging()
        instance._logger.setLevel(logging.INFO)
        instance._logger.info("Log level set to INFO")

    @staticmethod
    def disable_file_handler(debugging=False):
        """
        Disable logging to file and only log to stdout.

        :param debugging: If True, sets stream handler to DEBUG level; otherwise INFO level.
        """
        instance = UnmanicLogging()

        # Remove file handler if it exists
        if instance.file_handler:
            instance._logger.removeHandler(instance.file_handler)
            instance.file_handler = None
            instance._logger.info("File logging disabled. Logging only to stdout.")

        # Adjust stream handler level
        if instance.stream_handler:
            instance.stream_handler.setLevel(logging.DEBUG if debugging else logging.INFO)
            instance._logger.info(f"Stream logging set to {'DEBUG' if debugging else 'INFO'}")

    @staticmethod
    def update_stream_formatter(formatter):
        """
        Update the formatter of the stream handler.

        :param formatter: A logging.Formatter instance.
        """
        instance = UnmanicLogging()
        if instance.stream_handler:
            instance.stream_handler.setFormatter(formatter)
            instance._logger.info("Stream handler formatter updated.")
        else:
            instance._logger.warning("No stream handler found to update formatter.")

    @staticmethod
    def enable_remote_logging(endpoint, app_id):
        instance = UnmanicLogging()
        instance.remote_handler.configure_endpoint(endpoint, app_id)
        instance._logger.info("Remote logging enabled.")

    @staticmethod
    def disable_remote_logging():
        instance = UnmanicLogging()
        instance.remote_handler.configure_endpoint(None, None)
        instance._logger.info("Remote logging disabled.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.child_process.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     09 jULY 2025, (11:34 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import signal
import time
import queue
import threading

import psutil

from unmanic.libs.logs import UnmanicLogging

# Configure a global shared manager
_shared_manager = None

_active_plugin_pids = set()
_active_lock = threading.Lock()


def _register_pid(pid: int):
    with _active_lock:
        _active_plugin_pids.add(pid)


def _unregister_pid(pid: int):
    with _active_lock:
        _active_plugin_pids.discard(pid)


def kill_all_plugin_processes():
    """
    Terminate every plugin-spawned process (and its children)
    that’s still registered. Intended for use in atexit
    and tornado.autoreload hooks.
    """
    with _active_lock:
        pids = list(_active_plugin_pids)
        _active_plugin_pids.clear()

    for pid in pids:
        try:
            root = psutil.Process(pid)
        except psutil.NoSuchProcess:
            continue

        procs = root.children(recursive=True) + [root]

        # Ensure no processes are left in SIGSTOP
        for p in procs:
            try:
                # on Unix, unblock with SIGCONT
                p.send_signal(signal.SIGCONT)
            except Exception:
                pass
            try:
                # on all platforms psutil.resume() works if supported
                p.resume()
            except Exception:
                pass

        # Attempt graceful shutdown
        for p in procs:
            try:
                p.terminate()
            except psutil.NoSuchProcess:
                pass

        gone, alive = psutil.wait_procs(procs, timeout=3)

        # Finally, force kill any stragglers
        for p in alive:
            try:
                p.kill()
            except psutil.NoSuchProcess:
                pass
        psutil.wait_procs(alive, timeout=3)


def set_shared_manager(mgr):
    """Called once at service startup to inject the shared Manager."""
    global _shared_manager
    _shared_manager = mgr


class PluginChildProcess:
    def __init__(self, plugin_id, data):
        """
        data must include:
          - data['worker_log']              : list to which your child functions logs go
          - data['command_progress_parser'] : callable(line_text, pid=None, proc_start_time=None, unset=False)
        """
        self.logger = UnmanicLogging.get_logger(
            name=f'Plugin.{plugin_id}.{__class__.__name__}'
        )
        self.data = data
        if _shared_manager is None:
            raise RuntimeError("PluginChildProcess must be initialized after shared Manager is set")
        self.manager = _shared_manager
        self._log_q = self.manager.Queue()
        self._prog_q = self.manager.Queue()
        self._proc = None
        self._term_lock = threading.Lock()

    def run(self, target, *args, **kwargs):
        """
        Launch `target(*args, **kwargs, log_queue, prog_queue)` in its own process.
        Your `target` should accept two extra keyword args:
          log_queue  –> use log_queue.put(str) to emit log lines
          prog_queue –> use prog_queue.put(percentage:float) to emit progress
        """
        # Start child as before
        from multiprocessing import Process
        self._proc = Process(
            target=self._child_entry,
            args=(target, args, kwargs),
            daemon=True
        )
        self._proc.start()
        _register_pid(self._proc.pid)
        self.logger.info("Started child PID %s", self._proc.pid)

        # Register PID & start time with WorkerSubprocessMonitor
        parser = self.data.get('command_progress_parser')
        if callable(parser):
            try:
                parser(None, pid=self._proc.pid, proc_start_time=time.time())
            except Exception:
                self.logger.exception("Failed to register progress parser")

        # Drain logs, progress, watch exit
        success = self._monitor()

        # When the child process is done, unregister
        _unregister_pid(self._proc.pid)

        # Return success status
        return success

    def _child_entry(self, target, args, kwargs):
        """
        Runs inside the child process.
        Injects our two required queues into the call.
        """
        try:
            kwargs['log_queue'] = self._log_q
            kwargs['prog_queue'] = self._prog_q
            target(*args, **kwargs)
        except Exception:
            self.logger.exception("Exception in child target")

    def _monitor(self):
        """
        Parent loop: pull from log_q -> data['worker_log'],
                     pull from prog_q -> call parser(...)
        """
        exit_ok = False
        parser = self.data.get('command_progress_parser')

        while True:
            # 1) drain logs
            try:
                while True:
                    msg = self._log_q.get_nowait()
                    self.data['worker_log'].append(f"{msg}\n")
            except queue.Empty:
                pass

            # 2) drain progress updates
            try:
                while True:
                    pct = self._prog_q.get_nowait()
                    if callable(parser):
                        parser(str(pct))
            except queue.Empty:
                pass

            # 3) if the child exited, we’re done. Unset parser PID
            if not self._proc.is_alive():
                exit_ok = (self._proc.exitcode == 0)
                if callable(parser):
                    # tell parser to unset its internal proc state
                    parser(None, unset=True)
                break

            # Add a short wait here to prevent CPU pinning
            time.sleep(0.1)

        return exit_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.service.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     06 Dec 2018, (7:21 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import argparse
import os
import queue
import signal
import time
import threading

import psutil

from unmanic import config, metadata
from unmanic.libs import libraryscanner, common, eventmonitor
from unmanic.libs.db_migrate import Migrations
from unmanic.libs.logs import UnmanicLogging
from unmanic.libs.scheduler import ScheduledTasksManager
from unmanic.libs.taskqueue import TaskQueue
from unmanic.libs.postprocessor import PostProcessor
from unmanic.libs.taskhandler import TaskHandler
from unmanic.libs.uiserver import FrontendPushMessages, UIServer
from unmanic.libs.foreman import Foreman


def init_db(config_path):
    # Set paths
    app_dir = os.path.dirname(os.path.abspath(__file__))

    # Set database connection settings
    database_settings = {
        "TYPE":                       "SQLITE",
        "FILE":                       os.path.join(config_path, 'unmanic.db'),
        "MIGRATIONS_DIR":             os.path.join(app_dir, 'migrations_v1'),
        "MIGRATIONS_HISTORY_VERSION": 'v1',
    }

    # Ensure the config path exists
    if not os.path.exists(config_path):
        os.makedirs(config_path)

    # Create database connection
    from unmanic.libs.unmodels.lib import Database
    db_connection = Database.select_database(database_settings)

    # Run database migrations
    migrations = Migrations(database_settings)
    migrations.update_schema()

    # Return the database connection
    return db_connection


class RootService:

    def __init__(self):
        self.threads = []
        self.run_threads = True
        self.db_connection = None

        self.developer = None
        self.dev_api = None

        self.logger = UnmanicLogging.get_logger(name=__class__.__name__)
        UnmanicLogging.metric("root_service_started")

        self.event = threading.Event()

        self._mgr = None

    def start_handler(self, data_queues, task_queue):
        self.logger.info("Starting TaskHandler")
        handler = TaskHandler(data_queues, task_queue, self.event)
        handler.daemon = True
        handler.start()
        self.threads.append({
            'name':   'TaskHandler',
            'thread': handler
        })
        return handler

    def start_post_processor(self, data_queues, task_queue):
        self.logger.info("Starting PostProcessor")
        postprocessor = PostProcessor(data_queues, task_queue, self.event)
        postprocessor.daemon = True
        postprocessor.start()
        self.threads.append({
            'name':   'PostProcessor',
            'thread': postprocessor
        })
        return postprocessor

    def start_foreman(self, data_queues, settings, task_queue):
        self.logger.info("Starting Foreman")
        foreman = Foreman(data_queues, settings, task_queue, self.event)
        foreman.daemon = True
        foreman.start()
        self.threads.append({
            'name':   'Foreman',
            'thread': foreman
        })
        return foreman

    def start_library_scanner_manager(self, data_queues):
        self.logger.info("Starting LibraryScannerManager")
        library_scanner_manager = libraryscanner.LibraryScannerManager(data_queues, self.event)
        library_scanner_manager.daemon = True
        library_scanner_manager.start()
        self.threads.append({
            'name':   'LibraryScannerManager',
            'thread': library_scanner_manager
        })
        return library_scanner_manager

    def start_inotify_watch_manager(self, data_queues, settings):
        if eventmonitor.event_monitor_module:
            self.logger.info("Starting EventMonitorManager")
            event_monitor_manager = eventmonitor.EventMonitorManager(data_queues, self.event)
            event_monitor_manager.daemon = True
            event_monitor_manager.start()
            self.threads.append({
                'name':   'EventMonitorManager',
                'thread': event_monitor_manager
            })
            return event_monitor_manager
        else:
            self.logger.warn("Unable to start EventMonitorManager as no event monitor module was found")

    def start_ui_server(self, data_queues, foreman):
        self.logger.info("Starting UIServer")
        uiserver = UIServer(data_queues, foreman, self.developer)
        uiserver.daemon = True
        uiserver.start()
        self.threads.append({
            'name':   'UIServer',
            'thread': uiserver
        })
        return uiserver

    def start_scheduled_tasks_manager(self):
        self.logger.info("Starting ScheduledTasksManager")
        scheduled_tasks_manager = ScheduledTasksManager(self.event)
        scheduled_tasks_manager.daemon = True
        scheduled_tasks_manager.start()
        self.threads.append({
            'name':   'ScheduledTasksManager',
            'thread': scheduled_tasks_manager
        })
        return scheduled_tasks_manager

    def start_resource_logger(self):
        abort_flag = threading.Event()

        def log_resources():
            pid = os.getpid()
            proc = psutil.Process(pid)
            cpu_count = psutil.cpu_count(logical=True)
            start_time = time.time()

            while not self.event.is_set() and not abort_flag.is_set():
                try:
                    # Fetch CPU info
                    cpu_percent = proc.cpu_percent(interval=None)
                    normalised_cpu_percent = cpu_percent / cpu_count

                    # Fetch Memory info
                    mem_info = proc.memory_info()
                    rss_bytes = mem_info.rss
                    vms_bytes = mem_info.vms

                    # Calculate percentage of memory used relative to total system RAM
                    total_system_ram = psutil.virtual_memory().total
                    mem_percent = (rss_bytes / total_system_ram) * 100

                    # Calculate uptime in seconds
                    uptime = int(time.time() - start_time)

                    UnmanicLogging.metric("root_service_resources",
                                          pid=pid,
                                          uptime=uptime,
                                          cpu_percent=normalised_cpu_percent,
                                          mem_percent=mem_percent,
                                          rss_bytes=rss_bytes,
                                          vms_bytes=vms_bytes)
                except Exception as e:
                    self.logger.warning(f"Resource logging failed: {e}")
                    time.sleep(5)
                    continue

                time.sleep(5)  # Polling interval

        thread = threading.Thread(
            target=log_resources,
            name='RootServiceResourceLogger',
            daemon=True
        )
        thread.stop = abort_flag.set
        thread.start()
        self.threads.append({
            'name':   'RootServiceResourceLogger',
            'thread': thread
        })

    def initial_register_unmanic(self):
        from unmanic.libs import session
        s = session.Session(dev_api=self.dev_api)
        s.register_unmanic(s.get_installation_uuid())

    def start_threads(self, settings):
        # Create our data queues
        data_queues = {
            "library_scanner_triggers": queue.Queue(maxsize=1),
            "scheduledtasks":           queue.Queue(),
            "inotifytasks":             queue.Queue(),
            "progress_reports":         queue.Queue(),
            "frontend_messages":        FrontendPushMessages(),
        }

        # Clear cache directory
        self.logger.info("Clearing previous cache")
        common.clean_files_in_cache_dir(settings.get_cache_path())

        self.logger.info("Starting all threads")

        # Register installation
        self.initial_register_unmanic()

        # Setup job queue
        task_queue = TaskQueue(data_queues)

        # Setup post-processor thread
        self.start_post_processor(data_queues, task_queue)

        # Start the foreman thread
        foreman = self.start_foreman(data_queues, settings, task_queue)

        # Start new thread to handle messages from service
        self.start_handler(data_queues, task_queue)

        # Start scheduled thread
        self.start_library_scanner_manager(data_queues)

        # Start inotify watch manager
        self.start_inotify_watch_manager(data_queues, settings)

        # Start new thread to run the web UI
        self.start_ui_server(data_queues, foreman)

        # Start new thread to run the scheduled tasks manager
        self.start_scheduled_tasks_manager()

        # Start main thread resource logger
        self.start_resource_logger()

    def stop_threads(self):
        self.logger.info("Stopping all threads")
        self.event.set()
        for thread in self.threads:
            self.logger.info("Sending thread {} abort signal".format(thread['name']))
            thread['thread'].stop()
        for thread in self.threads:
            self.logger.info("Waiting for thread {} to stop".format(thread['name']))
            thread['thread'].join(10)
            self.logger.info("Thread {} has successfully stopped".format(thread['name']))
        self.threads = []

    def sig_handle(self, signum, frame):
        self.logger.info("Received {}".format(signum))
        self.stop()

    def stop(self):
        self.run_threads = False

    def run(self):
        # Init the TaskDataStore and PluginChildProcess
        import tornado.autoreload
        from multiprocessing import Manager
        import atexit
        from unmanic.libs.task import TaskDataStore
        from unmanic.libs.unplugins.child_process import kill_all_plugin_processes, set_shared_manager
        # Init a shared manager
        self._mgr = Manager()
        # Ensure Manager shuts down on process exit or tornado autoreload (dev mode)
        atexit.register(self._mgr.shutdown)
        tornado.autoreload.add_reload_hook(self._mgr.shutdown)
        # Ensure any PluginChildProcess shuts down on process exit or tornado autoreload (dev mode)
        atexit.register(kill_all_plugin_processes)
        tornado.autoreload.add_reload_hook(kill_all_plugin_processes)
        # Replace the in-process dicts with manager proxies
        TaskDataStore._runner_state = self._mgr.dict()
        TaskDataStore._task_state = self._mgr.dict()
        # Set the shared manager for PluginChildProcess
        set_shared_manager(self._mgr)

        # Init the configuration
        settings = config.Config()

        # Init the database
        self.db_connection = init_db(settings.get_config_path())

        # Start all threads
        self.start_threads(settings)

        # Watch for the term signal
        if os.name == "nt":
            while self.run_threads:
                try:
                    time.sleep(1)
                except (KeyboardInterrupt, SystemExit) as e:
                    break
        else:
            signal.signal(signal.SIGINT, self.sig_handle)
            signal.signal(signal.SIGTERM, self.sig_handle)
            while self.run_threads:
                signal.pause()
                time.sleep(.5)

        # Received term signal. Stop everything
        self.stop_threads()
        self.db_connection.stop()
        while not self.db_connection.is_stopped():
            time.sleep(.5)
            continue
        self.logger.info("Exit Unmanic")


def main():
    parser = argparse.ArgumentParser(description='Unmanic')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=metadata.read_version_string('long')))
    parser.add_argument('--manage_plugins', action='store_true',
                        help='manage installed plugins')
    parser.add_argument('--dev',
                        action='store_true',
                        help='Enable developer mode')
    parser.add_argument('--dev-api', nargs='?',
                        help='Enable development against another unmanic support api')
    parser.add_argument('--port', nargs='?',
                        help='Specify the port to run the webserver on')
    # parser.add_argument('--unmanic_path', nargs='?',
    #                    help='Specify the unmanic configuration path instead of ~/.unmanic')
    args = parser.parse_args()

    # Configure application from args
    settings = config.Config(
        port=args.port,
        unmanic_path=None
    )

    if args.manage_plugins:
        # Init the DB connection
        db_connection = init_db(settings.get_config_path())

        # Run the plugin manager CLI
        from unmanic.libs.unplugins.pluginscli import PluginsCLI
        plugin_cli = PluginsCLI()
        plugin_cli.run()

        # Stop the DB connection
        db_connection.stop()
        while not db_connection.is_stopped():
            time.sleep(.2)
            continue
    else:
        # Run the main Unmanic service
        service = RootService()
        service.developer = args.dev
        service.dev_api = args.dev_api
        service.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.file_browser_api.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     11 Apr 2021, (7:06 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import os

from unmanic.libs.uiserver import UnmanicDataQueues
from unmanic.webserver.api_v1.base_api_handler import BaseApiHandler


class ApiFilebrowserHandler(BaseApiHandler):
    name = None
    params = None
    unmanic_data_queues = None

    routes = [
        {
            "supported_methods": ["POST"],
            "call_method":       "fetch_directory_listing",
            "path_pattern":      r"/api/v1/filebrowser/list",
        },
    ]

    def initialize(self, **kwargs):
        self.name = 'file_browser_api'
        self.params = kwargs.get("params")
        udq = UnmanicDataQueues()
        self.unmanic_data_queues = udq.get_unmanic_data_queues()

    def set_default_headers(self):
        """Set the default response header to be JSON."""
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def get(self, path):
        self.action_route()

    def post(self, path):
        self.action_route()

    def fetch_directory_listing(self, *args, **kwargs):
        current_path = self.get_argument('current_path')
        list_type = self.get_argument('list_type') if self.get_body_arguments('list_type') else "all"

        path_data = self.fetch_path_data(current_path, list_type)

        self.write(json.dumps(path_data))

    def fetch_path_data(self, current_path, list_type="directories"):
        """
        Returns an object filled with data pertaining to a particular path

        :param current_path:
        :param list_type:
        :return:
        """
        directories = []
        files = []
        if list_type == "directories" or list_type == "all":
            directories = self.fetch_directories(current_path)
        if list_type == "files" or list_type == "all":
            files = self.fetch_files(current_path)
        path_data = {
            "current_path": current_path,
            "list_type":    list_type,
            "directories":  directories,
            "files":        files,
            "success":      True,
        }
        return path_data

    def fetch_directories(self, path):
        """
        Fetch a list of directory objects based on a given path

        :param path:
        :return:
        """
        results = []
        if os.path.exists(path):
            # check if this is a root path or if it has a parent
            parent_path = os.path.join(path, '..')
            if os.path.exists(parent_path) and os.path.abspath(parent_path) != path:
                # Path has a parent, Add the double dots
                results.append(
                    {
                        "name":      "..",
                        "full_path": os.path.abspath(parent_path),
                    }
                )
            for item in sorted(os.listdir(path)):
                abspath = os.path.abspath(os.path.join(path, item))
                if os.path.isdir(abspath):
                    results.append(
                        {
                            "name":      item,
                            "full_path": abspath,
                        }
                    )
        else:
            # Path doesn't exist!
            # Just return the root dir as the first directory option
            results.append(
                {
                    "name":      "/",
                    "full_path": "/",
                }
            )
        return results

    def fetch_files(self, path):
        """
        Fetch a list of file objects based on a given path

        :param path:
        :return:
        """
        results = []
        if os.path.exists(path):
            for item in sorted(os.listdir(path)):
                abspath = os.path.abspath(os.path.join(path, item))
                if os.path.isfile(abspath):
                    results.append(
                        {
                            "name":      item,
                            "full_path": abspath,
                        }
                    )
        return results
